/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#define AGENT_C
#include "bf_agent.h"

int     agent_re_exec = 0;
int     win_sock = 0;
char *  mapStr = NULL;
char *  defMapStr = NULL;
int     ifd = 0;
int     ofd = 1;
char *  trapCommand = NULL;
char *  clearcase_view = NULL;
int     verbose = 1;
FILE *  log_file = NULL;
char *  bf_debug = NULL;
int     agent_debug_flag = 0;
int     heartbeat_interval = 15;
char *  sourcelist = NULL;

bf_command_state_t command_state;

extern int did_hello;
extern ssl_state_t ssl_state;

extern const char *AGENT_VERSION;
extern const char *SUPPORTED_ENV_ACTIONS;

bf_command_t commands[] = {
	{ CMD_EXEC,     &FuncExec,      CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ CMD_EXPAND,   &FuncExpand,    CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ CMD_READ,     &FuncRead,      CMD_FLAG_CC | CMD_FLAG_ENV },
	{ CMD_WRITE,    &FuncWrite,     CMD_FLAG_CC | CMD_FLAG_ENV },
	{ CMD_MKDIR,    &FuncMkdir,     CMD_FLAG_CC | CMD_FLAG_ENV },
	{ CMD_RMDIR,    &FuncRmdir,     CMD_FLAG_CC | CMD_FLAG_ENV },
	{ CMD_UNLINK,   &FuncUnlink,    CMD_FLAG_CC | CMD_FLAG_ENV },
	{ CMD_PING,     &FuncPing,      CMD_FLAG_CC | CMD_FLAG_ENV | CMD_FLAG_PTY },
	{ CMD_SYSCONF,	&FuncSysConf,   CMD_FLAG_CC | CMD_FLAG_NO_REEXEC },
	{ CMD_SHOW,     &FuncDumpConf,  CMD_FLAG_NO_REEXEC },
	{ NULL, NULL, 0 }
};


/* 
 *  function to output a message of some sort. 
 *  used while reading the request from stdin. 
 */
static int agent_send_message(int code, const char *message) {
	char buf[128];
	int len = sprintf(buf, "%d ", code);
	len += format_message(&buf[len], message, AGENT_VERSION);
	buf[len++] = '\n';
	buf[len] = '\0';
	return io_puts(buf);
}

/* Become the requested user, if defined.  We also store the credentials,
 * for use later when we map drives.
 */
char *auth_user = NULL;
char *auth_pass = NULL;

static int agent_authenticate(void)
{
	int rc;
	const char *username = request_get(TAG_USERNAME);
	const char *password = request_get(TAG_ENCPASS);
	const char *fail_reason = NULL;
	char *saved_debug = bf_debug;
#ifdef WIN32
	char    *user = NULL;
#endif

	BFTRACE("agent_authenticate");

	/* For the duration of this function, we suppress client-requested
	 * debug information, as it would constitute an information leak,
	 * and thus a security risk.
	 */
	if (NULL != bf_debug) {
		BFTRACE("AUTH disabled remote debug trace");
		bf_debug = NULL;
	}

#ifdef WIN32
	/* On windows, we first use a temporary buffer to convert / to \. */
	if (NULL != username) {
		char *p;
		user = strdup(username);
		if (NULL != (p = strchr(user, '/')))
			*p = '\\';
		username = user;
	}
#endif
	if (NULL == password) { password = request_get(TAG_PASSWORD); }

	/* If we authenticated before, allow anything except an attempt to
	 * use a different username.
	 */
	if (auth_user) {
		if ( NULL == username
				|| '\0' == *username
				|| 0 == strcmp(username, auth_user) )
		{
			BFTRACE("AUTH ok (already auth'd)");
			rc = SU_OK;
		} else {
			BFTRACE("AUTH failed (user change)");
			rc = SU_ILLEGAL;
		}
	} else {
		BFTRACE("AUTH normal");
		rc = platform_set_user(username, password);
	}

	switch (rc) {
		case SU_OK:
			if (NULL == auth_user && NULL != username) {
				auth_user = strdup(username);
				if (NULL != password) { auth_pass = strdup(password); }
			}
			send_msg("AUTH", "AuthOk", "s", auth_user);
			rc = 0;
			break;

		case SU_OK_NOACCOUNT:
			rc = 0;
			break;

		case SU_ILLEGAL:
			fail_reason = "AuthIllegal";
			break;

		case SU_UNKNOWNUSER:
			fail_reason = "AuthUnknownUser";
			break;

		case SU_ACCOUNTINVALID:
			fail_reason = "AuthAccountInvalid";
			break;

		case SU_SETUSERFAILED:
			fail_reason = "AuthSetUserFailed";
			break;

		case SU_USERNOTFOUND:
			fail_reason = "AuthUserNotFound";
			break;

		case SU_PAMSESSIONERROR:
			fail_reason = "AuthSessionError";
			break;

		case SU_NOCRED:
			fail_reason = "AuthNoCredentials";
			break;

		case SU_NOSESSION:
			fail_reason = "AuthNoSession";
			break;

		default:
			fail_reason = "AuthSystemError";
			break;
	}

	if (rc != 0) {
		if (NULL != username) {
			send_msg("AUTH", "AuthFail", "ks",
					fail_reason, username);
		} else {
			send_msg("AUTH", "AuthFail", "kk",
					fail_reason, "AuthNoUser");
		}
	}

#ifdef WIN32
	/* On Windows, free the temporary buffer we used for
	 * converting / to \.
	 */
	if (NULL != user) { free(user); }
#endif

	/* Restore bf_debug state */
	if (NULL != saved_debug) {
		bf_debug = saved_debug;
		BFTRACE("AUTH re-enabled remote debug trace");
	}
	return rc;
}

void ClobberPasswords(void) {
	BFTRACE("ClobberPasswords");
	if (auth_pass) {
		memset(auth_pass, 0, strlen(auth_pass));
		free(auth_pass);
		auth_pass = 0;
	}
	if (mapStr) memset(mapStr, 0, strlen(mapStr));
	request_del_creds();
}

static void agent_process_command(bf_command_t *cmd)
{
	bf_command_state_t *state = &command_state;
	const char *bfconf_file = bfconf_get_path();
	char *env = request_get(TAG_ENV);

	int ret = 0;
#ifndef WIN32
	int pty = 0;
#endif

	activity_write(BFHERE "agent_process_command cmd=[%s]", cmd->cmd);

	memset(state, 0, sizeof(*state));
	state->mfd = state->sfd = -1;

	/* Turn on debugging if it was requested. */
	if (!bf_debug && request_has(TAG_DEBUG)) {
		bf_debug = calloc(1, BFBUFSIZ);
	}

#define DEBUG_WIN_REEXEC 0
#ifdef WIN32
	/* Authenticate, unless in a re-exec (Windows only) */
	if (agent_re_exec) {
		activity_write(BFHERE "AUTH disabled (agent_re_exec=%d)", agent_re_exec);
	} else {
		ret = agent_authenticate();
		if (0 != ret) {
			BFTRACE("AUTH failed");
			goto done;
		}
		BFTRACE("AUTH ok");

		if (!(cmd->flags & CMD_FLAG_NO_REEXEC) &&
				BFCONF_EXISTS(win_reexec_after_auth) &&
				(!IsOrigUser() || DEBUG_WIN_REEXEC))
		{
			BFTRACE("REEXEC starting");
			platform_reexec_server();
			return;
		}
	}

#else /* !WIN32 */
	/* Always authenticate normally (non-Windows only) */
	ret = agent_authenticate();
	if (0 != ret) {
		BFTRACE("AUTH failed");
		goto done;
	}
	BFTRACE("AUTH ok");
#endif /* WIN32 */

	/* build the environment */
	if (cmd->flags & CMD_FLAG_ENV) {
		if (NULL != sourcelist) {
			free(sourcelist);
			sourcelist = NULL;
		}
		BuildEnv(env, &sourcelist); 
		BFTRACE("ENV processed");
	} else {
		BFTRACE("ENV skipped");
	}

#ifndef WIN32
	/* Allocate pseudo-tty / pipe (non-Windows only).  We need a pty
	 * if the command always needs one (exec/expand) or if we are using
	 * CC views, as the commands to mount them will need one.
	 */
	pty = cmd->flags & CMD_FLAG_PTY;
	if (!pty && (cmd->flags & CMD_FLAG_CC)) {
		char *p;
		if (NULL != (p = platform_getenv("_CLEARCASE_VIEWS")) ||
			NULL != (p = platform_getenv("_CLEARCASE_VOBS")) )
		{
			platform_freeenv(p);
			pty = 1;
		}
	}
	if (pty) {
		BFTRACE("PTY allocating");
		if (0 != (ret = platform_openpty())) {
			BFTRACE("PTY failed");
			pty = 0;
			goto done;
		}
	} else {
		BFTRACE("PTY skipped");
	}
#endif /* WIN32 */


	/* Report any problems with the config file */
	if (0 != (ret = bfconf_get_errno())) {
		send_msg("EXEC", "ConfFail", "si", bfconf_file, ret);
	}


	/*
	 * Dispatch to the command.
	 */
	if (bf_debug) { request_dump(); }

	activity_write(BFHERE "invoking implementation for [%s]", cmd->cmd);
	ret = (*cmd->func)();

done:
	/* cleanup */
	ClobberPasswords();

#ifndef WIN32
	/* Release pseudo-tty / pipe (non-Windows only) */
	if (pty) {
		BFTRACE("PTY release");
		/* Trap SIGHUP */
		platform_signals_catch("bfagent");
		if (state->mfd != -1) {
			close(state->mfd);
			state->mfd = -1;
		}
		if (state->sfd != -1) {
			close(state->sfd);
			state->sfd = -1;
		}
		platform_signals_release();
	}
#endif /* WIN32 */
	send_result(ret);
}


/* 
 * Process a request block. this is basically a hash
 * of key=value pairs.
 */
int agent_process_request(void)
{
	const char * cmd;
	int          curr_cmd;

	BFTRACE("agent_process_request");

	if (NULL == (cmd = request_get(TAG_COMMAND))) {
		BFTRACE("No command (FATAL)");
		agent_send_message(CODE_NOCOMMAND, "RequestNoCommand");
		return 0;
	}

	/* locate the implementation of this command */
	for (curr_cmd = 0; commands[curr_cmd].cmd; ++curr_cmd) {
		if (0 == strcmp(commands[curr_cmd].cmd, cmd)) {
			/* Found a match; dispatch it! */
			agent_process_command(&commands[curr_cmd]);
			return 1;
		}
	}

	BFTRACE("Unknown command (FATAL)");
	agent_send_message(CODE_BADCOMMAND, "RequestInvalidCommand");
	return 0;
}


static const char *TOKENS_EOL = "\r\n";
static const char *TOKENS_SPACE = " \t\r\n";

static void bfagent_main_loop(void) {
	char *bf_rptr = NULL, *bf_sptr = NULL;
	char *buf = malloc(BFBUFSIZ), *tag, *params;
	const char *p;
	char empty = '\0';
	int traced_env = 0;
	int traced_stream = 0;
	int traced_userdata = 0;

	/*  read stdin, firing off requests as we get them */
	while (NULL != io_gets(buf, BFBUFSIZ)) {
		/*  tokenize out the first word in the buffer. */
		/*  this will be the tag word or command.      */
		bf_strtok(NULL, NULL, NULL, &bf_sptr);
		tag = bf_strtok(buf, TOKENS_SPACE, &bf_rptr, &bf_sptr);
		tag = LowerString(tag);
		if (NULL == tag || '\0' == *tag) { continue; }

		/* Discard lines with excessively long tags. */
		if (strlen(tag) > MAX_TAG_LEN) {
			BFTRACE("Tag too long (FATAL)");
			io_ayt();
			break;
		}

		/* Time to quit? */
		if (strcmp(tag, TAG_QUIT) == 0) {
			BFTRACE("tag: quit");
			break;
		}

		/* Set the language */
		if (strcmp(tag, TAG_LANG) == 0) {
			char *lang = bf_strtok(buf, TOKENS_SPACE, &bf_rptr, &bf_sptr);
			platform_set_language(lang);
			request_del(tag);
			request_put(tag, lang);
			BFTRACE("tag: lang");
			continue;
		}

		/* Request is complete, so process it. */
		if (strcmp(tag, TAG_GO) == 0) {
			BFTRACE("tag: go");
			if (!agent_process_request()) { break; }
			log_close();
			env_clear();
			request_new();
			traced_env = traced_stream = traced_userdata = 0;
			continue;
		}

		/* Make sure we don't need more entries than are allowed. */
		if (request_size() > MAX_REQUEST_TAGS && !request_has(tag)) {
			BFTRACE("Too many tags (FATAL)");
			io_ayt();
			break;
		}

		/*  at this point we'll need any trailing args in the  */
		/*  input buffer. if there are none, set an empty string */
		/*  rather than a null pointer to make life easier later on. */
		params = bf_strtok(NULL, TOKENS_EOL, &bf_rptr, &bf_sptr);
		if (!params) { params = &empty; }

		/* the userdata tag is hex encoded so we can transport binary data */
		if (strcmp(tag, TAG_USERDATA) == 0) {
			size_t len = strlen(params) >> 1;
			unsigned char *value = calloc(1, len);
			hex_decode(value, params);
			request_put_binary(tag, value, len);
			if (!traced_userdata) {
				BFTRACE("tag: userdata");
				traced_userdata = 1;
			}
			continue;
		}

		/* The stream tag is used to record file data to a temporary file.
		 * If the file has not been opened yet, then we need to create it
		 * now.
		 */
		if (strcmp(tag, TAG_STREAM) == 0 || strcmp(tag, TAG_DATA) == 0) {
			size_t len = strlen(params) >> 1;
			unsigned char *value = calloc(1, len);
			hex_decode(value, params);
			request_put_stream(TAG_STREAM, value, len);
			if (!traced_stream) {
				BFTRACE("tag: stream");
				traced_stream = 1;
			}
			continue;
		}

		/* the encpass tag is bf_encrypt'd so the password
		 * will not have to be sent in plaintext.
		 */
		if (strcmp(tag, TAG_ENCPASS) == 0) {
			char *password = NULL;
			request_del(tag);
			decrypt_pwd(&password, params);
			if (NULL != password) {
				request_put(tag, password);
			}
			continue;
		}

		if (strcmp(tag, TAG_ENV) == 0) {
			/* Discard any settings we don't actually understand. */
			if (*params == '\0' || NULL == strchr(SUPPORTED_ENV_ACTIONS, *params)) {
				BFTRACE("tag: env (unknown type)");
				continue;
			}
			if (!traced_env) {
				BFTRACE("tag: env");
				traced_env = 1;
			}
		}

		/* Check for a few invalid tags that would come from
		 * port scanners that think we speak HTTP.
		 */
		if (0 == strcmp(tag, "get") ||
				0 == strcmp(tag, "post") ||
				0 == strcmp(tag, "help") ||
				0 == strcmp(tag, "options"))
		{
			io_ayt();
			break;
		}

		/* For anything else, we just save the value.  If we
		 * have already seen this tag, then append to it with
		 * this content, using a newline to separate it.  But
		 * first, make sure it is just text.  If somebody is
		 * sending up binary garbage, just hang up on them.
		 */
		p = tag;
		while ('\0' != *p && isalpha((int)(*p))) { ++p; }
		if ('\0' != *p) {
			io_ayt();
			break;
		}

		activity_write(BFHERE "tag: %s", tag);
		request_put(tag, params);
	}

	request_destroy();
	bf_strtok(NULL, NULL, NULL, &bf_sptr);
	free(buf);
}



/*  
 *  process client requests
 */
int bfagent(void) {
	char    *buf;

	did_hello = 0;

	if (NULL != (buf = bfconf_get("debug_pause", NULL))) {
		platform_sleep(atoi(buf));
		free(buf);
	}

	/* If we're supporting telnet, then enable SO_OOBINLINE */
	if (!BFCONF_EXISTS(disable_telnet_support)) {
		platform_protocol_oobinline();
	}

	/* Initialize locale and lang settings */
	{
		char *default_locale = bfconf_get("locale", NULL);
		if (NULL != default_locale) {
			platform_setenv("LANG", default_locale, 1);
			free(default_locale);
		}
	}

	platform_setlocale("");
	platform_set_language("");

	if (NULL != (buf = bfconf_get("heartbeat_interval", NULL))) {
		heartbeat_interval = atoi(buf);
		free(buf);
	}

	if (NULL != (buf = bfconf_get("command_output_cache", NULL))) {
		int size = atoi(buf);
		if (size < 2048) { size = 2048; }
		free(buf);
		send_cache_init(size);
	}

	defMapStr = bfconf_get("map", NULL);
	request_new();

	if (agent_re_exec) {
		return platform_reexec_client();
	} else if (NULL != start_SSL() || SS_OK == ssl_state) {
		bfagent_main_loop();
		stop_SSL();
	}

	/* release memory we don't need. */
	if (BFCONF_EXISTS(disable_cleanup_during_exit)) {
		BFTRACE("cleanup disabled");
	} else {
		BFTRACE("cleanup");
		platform_release_credentials();
		free(buf);
		request_destroy();
		platform_freeenv(clearcase_view);
	}
	return 0;
}




