/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved.
IBM, Inc. owns the copyright and other intellectual
property rights in this software. Duplication or use of the
Software is not permitted except as expressly provided in a
written agreement between your company and IBM, Inc.
*/

#define BF_USE_REGEX

#include "bf_agent.h"

extern const char *AGENT_VERSION;

static char *valid_var_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789-?";

static hash_ctx *env_hash_ctx = NULL;
static hash_table *env_ht_hidden = NULL;

static char *copy_var_name(const char *name) {
	char *p = strchr(name, '=');
	int len = (NULL != p) ? (p - name) : (int)strlen(name);

	p = malloc(len + 1);
	memcpy(p, name, len);
	p[len] = '\0';
	return p;
}

void env_clear(void) {
	if (NULL != env_ht_hidden) {
		hash_destroy(env_ht_hidden);
		env_ht_hidden = NULL;
	}
}

int env_is_hidden(const char *name) {
	if (NULL != env_ht_hidden) {
		char *p = copy_var_name(name);
		int ret = hash_contains_key_str(env_ht_hidden, p);
		free(p);
		return ret;
	}
	return 0;
}

static const char *env_filtered_value(const char *var, const char *value)
{
	return env_is_hidden(var) ? "*****" : value;
}



/* Spits out data blocks that report on the locale status of the agent.
 * This is called automatically by BuildEnv.  Anything like ping that
 * wants this information displayed but does not call BuildEnv should
 * call this directly, instead.
 */
void env_report_locale(void) {
	char *want = platform_getenv("LANG");
	const char *got = platform_setlocale(want);

	if (NULL != want) {
		if (NULL == got) {
			int err = errno;
			got = platform_setlocale(NULL);
			send_msg("EXEC", "LocaleFail", "si", want, err);
		}
		platform_freeenv(want);
	}

	if (NULL != got) {
		send_msg("EXEC", "Locale", "s", got);
	}
}





/*
 * This checks to see if you are assigning to the "PATH" variable on
 * Unix and if so converts any Windows path separators (';') to Unix
 * path separators (':').  The intent behind this is that you might
 * have variables that you set differently on Windows and Unix, such
 * as JAVA_HOME and ANT_HOME, but at some point need to add both of
 * them to the PATH regardless of platform.  You could then create an
 * APPEND type entry that used
 *     $JAVA_HOME/bin;$ANT_HOME/bin
 * as the value, and because of this function, it should work equally
 * well on both platforms.
 *
 * This approach is unsatisfactory because:
 * - It is "magic" -- that is, although it is convenient, it is
 *   something that you wouldn't normally expect to happen.  The
 *   Principle of Least Surprise is generally a better strategy, as
 *   excessive magic tends to lead to bugs.
 * - It only happens for PATH; it does not work for CLASSPATH or
 *   any other PATH-like variable.
 * - This would prevent you from successfully using a PATH that
 *   includes a directory with the ';' character in it.  It is unlikely
 *   that one would be needed, but it is perfectly legal in Unix
 *   filenames, so not allowing it is an unnecessary constraint.
 * - The current implementation is one-way.  That is, if you specify
 *   a Windows-style separator on Unix, it gets changed, but Unix-style
 *   separators are preserved if used on Windows.
 * - Trying to extend this behavior to work the other direction
 *   would exacerbate the problem, as ':' is commonly part of
 *   Windows pathnames, so we would have to distinguish between
 *   drive separators and pathname separators heuristically.
 * - There was previously no way to turn it off
 * - The same functionality is better and more portably provided by
 *   using multiple APPEND entries.
 *
 * The only reason this code remains is for backward compatibility.  We
 * should discourage its (ab)use and eventually remove it.  This is now
 * guarded by a bfagent.conf parameter.
 */
static void check_path_style(const char *var, char *value)
{
	if (BFCONF_EXISTS(enable_deprecated_path_mapping)
			&& 0 == strcmp(var, "PATH")
			&& STYLE_UNIX == platform_style())
	{
		char *p = value;
		while (NULL != ( p = strchr(p, ';') )) {
			*p++ = ':';
		}
	}
}



/* Implements the (S)et and Set If (N)ot Set actions directly.
 *
 * Other actions should also call this internally to make sure
 * that things like _MAP are handled properly regardless of
 * the action type.
 *
 * overwrite - use 0 to disable overwriting; otherwise, the
 * 		variable will be set in all cases
 * report - use 0 to suppress the send_msg calls; otherwise,
 * 		the depend only on the global "verbose" setting
 * Return Value: 0 if the variable was not set because overwriting
 * 		is disabled; otherwise, 1
 */
static int env_set(
		const char *var,
		const char *value,
		int overwrite,
		int report)
{
	const char *msg = "EnvSet";
	char *v = env_expand_var(value, 0);

	if (!verbose) { report = 0; }

	if (!overwrite) {
		char *curr = platform_getenv(var);
		if (NULL != curr) {
			if (verbose) {
				send_msg("SET", "EnvAlreadySet", "sss",
						var, env_filtered_value(var, curr), value);
			}
			platform_freeenv(curr);
			return 0;
		}
		msg = "EnvSetNew";
	}

	/* convert a windows path to unix format */
	check_path_style(var, v);
	if (strcmp(var, "_MAP") == 0) {
		if (NULL != mapStr) { free(mapStr); }
		mapStr = strdup(v);
	} else if (strcmp(var, "_TRAP") == 0) {
		if (NULL != trapCommand) { free(trapCommand); }
		trapCommand = strdup(v);
	} else if (strcmp(var, "_BFAGENTDEBUG") == 0) {
		if (!bf_debug) {
			bf_debug = calloc(1, BFBUFSIZ);
			if (report) {
				send_msg("DBG", "EnvDebug", NULL);
			}
		}
	} else {
		platform_setenv(var, v, 1);
	}
	if (report) {
		send_msg("SET", msg, "ss", var, env_filtered_value(var, v));
	}
	free(v);
	return 1;
}



static void env_set_hidden(const char *name, const char *value) {
	if (NULL == env_ht_hidden) {
		if (NULL == env_hash_ctx)
			env_hash_ctx = hash_ctx_new();
		env_ht_hidden = hash_new(env_hash_ctx, 11); 
	}
	hash_put_str(env_ht_hidden, copy_var_name(name), NULL);
	env_set(name, value, 1, 0);
	if (verbose) {
		send_msg("SET", "EnvHidden", "s", name);
	}
}



static void env_set_source(const char *value, char **src) {
	char *p, *s, *v;
	size_t vlen;

	/* Discard trailing whitespace */
	v = env_expand_var(value, 0);
	vlen = strlen(v);
	while (vlen > 0) {
		switch (v[vlen-1]) {
			case '\t': case '\n': case '\f': case '\r': case ' ':
				v[(int)(--vlen)] = '\0';
				continue;
		}
		break;
	}

	if (vlen == 0) {
		free(v);
		return;
	}

	/* Discard leading whitespace */
	for (p=v; *p; ++p) {
		switch (*p) {
			case '\t': case '\n': case '\f': case '\r': case ' ':
				--vlen;
				continue;
		}
		break;
	}

	/* create a list of pre-exec source commands */
	if (NULL == *src) {
		s = *src = calloc(1, vlen + 2);
	} else {
		size_t len = strlen(*src);
		s = *src = realloc(*src, len + vlen + 2);
		s += len;
	}

	memcpy(s, p, vlen);
	s += vlen;
	*s++ = '\n';
	*s++ = '\0';

	if (verbose) {
		send_msg("SRC", "EnvSource", "s", v);
	}
	free(v);
}


static void env_set_date(const char *var, const char *curr) {
	char *newValue;

	while (isspace(( int)(*curr) )) { ++curr; }
	if ('\0' == *curr) {
		if (verbose) {
			send_msg("SET", "EnvCmdNoFormat", "s",
					".date");
		}
		return;
	}

	newValue = expandDate(curr);
	platform_setenv(var, newValue, 1);
	if (verbose) {
		send_msg("SET", "EnvSetIndirect", "ssss",
				var, newValue, ".date", curr);
	}
	free(newValue);
}



static void env_set_gmdate(const char *var, const char *curr) {
	char *newValue;

	while (isspace(( int)(*curr) )) { ++curr; }
	if ('\0' == *curr) {
		if (verbose) {
			send_msg("SET", "EnvCmdNoFormat", "s",
					".gmdate");
		}
		return;
	}

	newValue = expandGMDate(curr);
	platform_setenv(var, newValue, 1);
	if (verbose) {
		send_msg("SET", "EnvSetIndirect", "ssss",
				var, newValue, ".gmdate", curr);
	}
	free(newValue);
}



/* .substr "$STR" start [len] */
static void env_set_substr(const char *var, const char *args) {
	const char *err;
	char *curr, *newValue;
	int start, len = -1, vlen;
	char *cl = strdup(args);
	char **argv = getArgs(cl);

	if (NULL == argv) {
		err = "EnvSubstrNoArgs";
		goto done;
	}
	if (NULL == argv[0]) {
		err = "EnvSubstrNoSource";
		goto done;
	}

	start = (NULL != argv[1] && isdigit( (int)(*argv[1]) ))
			? atoi(argv[1])
			: -1;
	if (start < 0) {
		err = "EnvSubstrBadStartIdx";
		goto done;
	}

	if (NULL != argv[2]) {
		len = isdigit( (int)(*argv[2]) ) ? atoi(argv[2]) : -1;
		if (len < -1) {
			err = "EnvSubstrBadEndIdx";
			goto done;
		}
	}

	curr = env_expand_var(argv[0], 0);
	vlen = (int)strlen(curr) + 1;
	newValue = malloc(vlen);
	memcpy(newValue, curr, vlen);

	if (start > 0) {
		if (start < len) {
			memmove(curr, &curr[start], vlen - start);
			vlen -= start;
		} else {
			*curr = '\0';
			vlen = 0;
		}
	}

	if (len >= 0 && len < vlen) { curr[len] = '\0'; }
	if (verbose) {
		send_msg("SET", "EnvSubstr", "sssii",
				var, curr, newValue, start, len);
	}
	platform_setenv(var, curr, 1);
	free(newValue);
	free(curr);
	err = NULL;

done:
	freeArgs(argv);
	free(cl);
	if (verbose && NULL != err) {
		send_msg("SET", err, NULL);
	}
}


static void env_set_regex(const char *var, const char *args) {
	/* .regex /pattern/repl/ $value */
	/* .regex /pattern/repl/ value */
	bf_regex_t *srch = NULL;
	char *spat = NULL;
	char *repl = NULL;
	char *target = NULL;
	char *curr;

	while (isspace( (int)(*args) )) { ++args; }
	curr = strdup(args);

	if (0 != bf_regex_prep(&srch, &spat, &repl, &target, curr)) {
		if (verbose) {
			send_msg("SET", "EnvRegexSyntax", "ks",
					target, curr);
		}
	} else {
		char *newValue = bf_regex_apply(target, srch, repl);
		char *value = (NULL != newValue) ? newValue : target;

		platform_setenv(var, value, 1);
		if (verbose) {
			send_msg("SET", "EnvRegexApply", "sss",
					var, value, args);
		}
		if (NULL != newValue) { free(newValue); }

		bf_regex_free(srch);
		free(spat);
		free(repl);
	}

	free(curr);
	free(target);
}




/* Implements the (A)ppend and (P)repend actions */
static void env_set_concat(
		char cmd,
		const char *var,
		const char *value)
{
	char *v, *curr;

	v = env_expand_var(value, 0);
	curr = platform_getenv(var);

	check_path_style(var, v);

	if (NULL != curr) {
		/* Have an old value, so append/prepend as requested. */
		char *newValue = calloc(
				sizeof(*newValue),
				strlen(v) + strlen(curr) + 2);
		if (cmd == 'A') {
			format_message(newValue, "%s%c%s", curr, EnvPathSeparator(), v);
			if (verbose) {
				send_msg("SET", "EnvConcat", "sss",
						var, curr, value);
			}
		} else {
			format_message(newValue, "%s%c%s", v, EnvPathSeparator(), curr);
			if (verbose) {
				send_msg("SET", "EnvConcat", "sss",
						var, value, curr);
			}
		}
		platform_setenv(var, newValue, 1);
		free(newValue);
		platform_freeenv(curr);
	} else {
		/* No old value, so just set it normally. */
		platform_setenv(var, v, 1);
		if (verbose) {
			send_msg("SET", "EnvSet", "ss", var, v);
		}
	}

	free(v);
}





/* Everything in the switch statement below needs to appear in this list. */
const char *SUPPORTED_ENV_ACTIONS = "SAPNCUHE";
static void env_set_cmd(char cmd, const char *var, const char *value) {
	/*
	 * Everything in the switch statement below needs to appear in
	 * SUPPORTED_ENV_ACTIONS.  Otherwise, they will get discarded
	 * when they are received.
	 */
	switch (cmd) {
		case 'S':		/* set var to value */
			env_set(var, value, 1, 1);
			break;
		case 'A':		/* append value to var */
		case 'P':		/* prepend value to var */
			env_set_concat(cmd, var, value);
			break;
		case 'N':		/* set if not already set */
			env_set(var, value, 0, 1);
			break;
		case 'C':		/* clear the var */
			platform_setenv(var, "", 1);
			send_msg("SET", "EnvClear", "s", var);
			break;
		case 'U':		/* unset the var - no way to know if this works on all platforms. */
			platform_unsetenv(var);
			send_msg("SET", "EnvUnset", "s", var);
			break;
		case 'H':		/* set var to hidden value */
			env_set_hidden(var, value);
			break;
	}
}


/* The envs var is a string of the form "S:X:Y\nA:C:D\nP:E:F\nC:Z:\n".
 * It is modified in-place (and therefore destroyed) by this function.
 */
void BuildEnv(char *envs, char **src) {
	const char *locale;
	char dummy[1];
	char *cmd, *var, *value;
	char *bf_rptr = NULL;
	char *bf_sptr = NULL;

	if (NULL != src) { *src = NULL; }
	if (NULL == envs) {
		*dummy = '\0';
		envs = dummy;
	}

	/* If we have a locale, then make sure the LANG setting starts out
	 * matching it.
	 */
	locale = platform_setlocale("");
	if (NULL != locale) {
		platform_setenv("LANG", locale, 0);
	}

	/* set the agent version and OS platform */
	platform_setenv("BF_AGENT_VERSION", AGENT_VERSION, 1);
	send_msg("SET", "EnvSet", "ss",
			"BF_AGENT_VERSION", AGENT_VERSION);

	{
		const char *id = platform_id();
		platform_setenv("BF_AGENT_PLATFORM", id, 1);
		send_msg("SET", "EnvSet", "ss",
				"BF_AGENT_PLATFORM", id);
	}

	for (cmd = bf_strtok(envs, "\n", &bf_rptr, &bf_sptr);
			NULL != cmd;
			cmd = bf_strtok(NULL, "\n", &bf_rptr, &bf_sptr) )
	{
		var = strchr(cmd, ':');
		if (NULL == var) { continue; }
		*var++ = 0;
		value = strchr(var, ':');
		if (NULL == value) { continue; }
		*value++ = 0;

		if (0 == strncmp(var, "BF_", 3)) { value = CleanPath(value); }

		if (NULL != src && 0 == strcmp(var, ".source")) {
			env_set_source(value, src);
		} else if (strncmp(value, ".date", 5) == 0) {
			env_set_date(var, &value[5]);
		} else if (strncmp(value, ".gmdate", 7) == 0) {
			env_set_gmdate(var, &value[7]);
		} else if (strncmp(value, ".substr", 7) == 0) {
			env_set_substr(var, &value[7]);
		} else if (strncmp(value, ".regex", 6) == 0) {
			env_set_regex(var, &value[6]);
		} else {
			env_set_cmd(*cmd, var, value);
		}
	}
	bf_strtok(NULL, NULL, NULL, &bf_sptr);

	/* Pay attention to and report a change to the locale. */
	env_report_locale();

	if (!BFCONF_EXISTS(ignore_clearcase_view))
		clearcase_view = platform_getenv("CLEARCASE_VIEW");
}


/*
 * function to compare env strings in a platform dependant way.
 * In truth, this is just strncmp or strnicmp as appropriate.
 */
static int env_name_match(const char *str1, const char *str2, int len) {
	int lower = (platform_style() != STYLE_WINDOWS) ? 1 : 0;
	while (len-- > 0) {
		int c1 = (*str1++) & 0xFF;
		int c2 = (*str2++) & 0xFF;
		if (lower) {
			c1 = tolower(c1);
			c2 = tolower(c2);
		}
		if (c1 == 0) { return c2; }
		if (c1 != c2) { return c2 - c1; }
	}
	return 0;
}


/*
 * expand a string that has an environment variable in it.
 *
 * handle translation of windows and unix for commonality
 *
 */

#define UA_NONE 0
#define UA_WARN 1
#define UA_USE_EMPTY 2

char *env_expand_var_impl(
		const char *src,
		int ignore_escapes,
		int do_hidden_check)
{
	char	*p, *value, *remainder, **env, *q;
	char	*result;
	int 	varlen, envlen, parse_skip, did_subst, len;
	int		default_undef_action = UA_WARN;
	int		undef_action;
	char	**env_ptr;
	char    *src2;
	int		advance_pointer=0;

	if (!src) { return NULL; }

	/* Find out whether $VAR, ${VAR}, and %VAR% should expand as var name
	 * or in shell-compatible way.  For the time being, %VAR% will expand
	 * as blank instead of staying %VAR%, which is wrong...  That will be
	 * fixed later.
	 */
	if (BFCONF_EXISTS(shell_compatible_undef_vars))
		default_undef_action |= UA_USE_EMPTY;

	/* make a local copy of the source string */
	/* we use calloc/strcpy instead of strdup so we can realloc later */
	result = strdup(src);

	/* do windows env var conversion, %X% -> ${X} */
	for (p = result ; *p ; p++) {
		/* start of a var or just another character */
		if (*p != '%')
			continue;

		/* the % must preceed a valid variable */
		/* if it doesn't, skip over it and continue looking */
		if ((varlen = strspn(p+1, valid_var_chars)) < 1)
			continue;

		/* the terminating character has to be a % for this to be valid */
		if (*(p + varlen + 1) != '%')
			continue;

		/* it's a legal var, convert it to unix style */
		*p++ = 0;
		*(p + varlen) = '}';

		/* need to grow the string by one char, to insert the extra curly */
		src2 = strdup(p);
		parse_skip = strlen(result);
		result = realloc(result, parse_skip + strlen(src2) + 3);
		result[parse_skip++] = '$';
		result[parse_skip++] = '{';
		strcpy(&result[parse_skip], src2);
		free(src2);
		p = &result[parse_skip];
	}

	/* scan the result buffer for $ vars */
	/* p may get bumped off the end of the buffer so we check to see if that's the case */
	/* instead of checking *p as our for loop test. a segv safe test :) */
	parse_skip = 0;	/* how much to skip over on parser restarts */
	env_ptr = NULL;
	for (p = result ; p < &(result[strlen(result)]); ) {
		if (advance_pointer == 1) {
			p++;
		} else {
			advance_pointer = 1;
		}

		/* a single quote? */
		if (*p == '\'' && (q = strchr(p+1, '\'')) != NULL)
			p = q+1;

		/* an escaped character? */
		if (!ignore_escapes && *p == '\\' && platform_style() != STYLE_WINDOWS) {
			memmove(p, p+1, strlen(p));
			continue;
		}

		/* start of a variable ? */
		if (*p != '$')
			continue;

		/* if we have a $$, convert it to a single $ and skip over it */
		if (*(p+1) == '$') {
			memmove(p, p+1, strlen(p));
			continue;
		}

		/* at this point, p points to a $ */
		undef_action = default_undef_action;

		/* ${VAR} format */
		if (p[1] == '{') {
			/* is there a matching right curly after a valid variable name ? */
			if (0 != (varlen = strspn(p+2, valid_var_chars)) && p[varlen + 2] == '}') {
				/* Skip over the ${ */
				memmove(p, p+2, varlen);
				/* Eat the trailing } */
				memmove(p + varlen, p + varlen + 3, strlen(p + varlen + 3) + 1);
			}
			else {
				++p;
				continue;
			}
		}

		/* $[VAR] is like ${VAR}, but if a variable is undefined, it will always
		 * use "" for the value instead of "VAR", and it does not generate the
		 * warning message.
		 */
		else if (p[1] == '[') {
			undef_action = UA_USE_EMPTY;
			/* is there a matching right bracket after a valid variable name ? */
			if ((varlen = strspn(p+2, valid_var_chars)) && *(p + varlen + 2) == ']') {
				/* Skip over the $[ */
				memmove(p, p+2, varlen);
				/* Eat the trailing ] */
				memmove(p + varlen, p + varlen + 3, strlen(p + varlen + 3) + 1);
			}
			else {
				++p;
				continue;
			}
		}

		/* $VAR format */
		else {
			if ((varlen = strspn(p+1, valid_var_chars))) {	/* find the length of this var */
				/* it's legal, clean it up - this eats the $ */
				memmove(p, p+1, strlen(p));
			}
			else {
				/* not a legal variable, skip over it */
				++p;
				continue;
			}
		}


		/* p points to a variable name and varlen says how long it is */
		parse_skip = p - result;	/* remember how much is already done to protect expanded $'s */
		q = calloc(1, varlen+1);
		memcpy(q, p, varlen);

		if (do_hidden_check && NULL != env_ht_hidden) {
			if (env_is_hidden(q)) {
				send_msg("SET", "EnvHiddenNoPreparse", "s", q);
				free(q);
				free(result);
				result = strdup(src);
				if (env_ptr != NULL) { freeBFenviron(env_ptr); }
				return result;
			}
		}

		did_subst = 0;
		for (env = env_ptr = getBFenviron() ; *env ; env++) {
			/* get the length of the env variable */
			envlen = strcspn(*env, "=");

			/* compare the variable name pointed to by p with the env variable name */
			if (envlen == varlen && env_name_match(p, *env, varlen) == 0) {
				value = strchr(*env, '=');
				value++;

				/* null the working string and point to the remainder after the variable */
				*p = 0;
				remainder = strdup(p + varlen);	/* make a copy of the stuff after the var name in the working string */
											/* because we're going to realloc the working string buffer shortly */
				/* rebuild the working string */
				len = strlen(result);
				result = realloc(result, len + strlen(value) + strlen(remainder) + 1);
				strcpy(&result[len], value);
				strcat(&result[len], remainder);
				free(remainder);
				p = result + parse_skip;
				advance_pointer=0;
				did_subst = 1;
				break;
			}
		}

		if (!did_subst) {
			/* Not $[VAR] format, so warn about undefined var */
			if (undef_action & UA_WARN) {
				send_msg("SET", "EnvWarnUndef", "s", q);
			}

			/* $[VAR] format or shell_compatible_undef_vars was set */
			if (undef_action & UA_USE_EMPTY) {
				memmove(p, p + varlen, strlen(p + varlen) + 1);
			}
		}
		free(q);
	}
	freeBFenviron(env_ptr);
	return result;
}


