/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"

/* Activity logging for this file is suppressed unless IO_DEBUG
 * is defined during compilation.
 */
#ifndef IO_DEBUG
#  undef activity_file
#  define activity_file NULL
#endif

ssl_state_t ssl_state = SS_OK;

#ifdef HAVE_SSL
/*================================================================
 * Compiled with OpenSSL
 *----------------------------------------------------------------*/

#ifdef WIN32
extern SOCKET hRemote;
# else
#  define hRemote ofd 
#endif
extern int win_sock;
extern int ofd;
extern int ifd;

SSL_CTX *ctx = NULL;
void *ssl = NULL;

static int verify_SSL_callback(int flag, X509_STORE_CTX *store) {
	int 	err;

	if (!flag) {
		err = X509_STORE_CTX_get_error(store);
		send_msg("SSL", "SSLErrorUnverifiedCert", "is",
				err, X509_verify_cert_error_string(err));
		return 0;
	}
	return 1;
}

ssl_state_t init_CTX(void) {
	STACK_OF(X509_NAME) *cert_names=NULL;
	const char *reason = NULL;
	char *ssl_key_location = NULL, *ssl_key_password = NULL;
	char *ssl_cert_location = NULL, *ssl_cert_password = NULL;
	char *ssl_ca_location = NULL, *ssl_ca_password = NULL;
	char *cipher_group = NULL, *cipher_override = NULL, *client_authentication = NULL;
	char *fips_enabled = NULL, *ssl_protocol = NULL;
	char buf[1024], err_buf[1024], *sbuf=NULL;
	int err=0;

	/* Paranoid... init_CTX shouldn't get called more than once... */
	if (ctx) { SS_FAILED; }

	/* Per docs, these never fail. */
	SSL_load_error_strings();
	ERR_load_crypto_strings();
	SSL_library_init();

	/* See if FIPS is enabled early on. */
	fips_enabled = bfconf_get("fips_enabled", "false");

	sprintf(buf, "FIPS enabled: %s", fips_enabled);
	BFTRACE(buf);

	#ifdef OPENSSL_FIPS
		if (fips_enabled != NULL && !strcmp("true", fips_enabled)) {
			if (!FIPS_mode()) {  /* Check if FIPS already enabled. */
				if (!FIPS_mode_set(1)) {
					reason = "SSLErrorFIPSEnablement";
					goto done;
				}
				BFTRACE("Successfully enabled FIPS.");
			}
		}
	#endif

	/* If no cert is defined, then we don't need ssl. */
	ssl_key_location = bfconf_get("ssl_key_location", NULL);
	if (NULL == ssl_key_location) { return SS_OK; }
	
	sprintf(buf, "Key location: %s", ssl_key_location);
	BFTRACE(buf);

	/* A PEM will need a password if encrypted. */
	ssl_key_password = getPasswordFromAgentConfig ("ssl_key_password");

	/* Get the cert. */
	ssl_cert_location = bfconf_get("ssl_cert_location", NULL);
    
	sprintf(buf, "Cert location: %s", ssl_cert_location);
	BFTRACE(buf);

	/* Probably not used, but just in case.. */
	ssl_cert_password = getPasswordFromAgentConfig ("ssl_cert_password");

	/* Get the SSL protocol. */
	ssl_protocol = bfconf_get("ssl_protocol", NULL); if (ssl_protocol == NULL) ssl_protocol = "TLSv1";
	#ifdef OPENSSL_FIPS
		if (FIPS_mode()) ssl_protocol = "TLSv1";
	#endif

	/* Create the SSL context with the desired protocol. */
	if (!strcmp(ssl_protocol, "TLSv1") || !strcmp(ssl_protocol, "TLS"))  
		ctx = SSL_CTX_new(TLSv1_server_method());
	else ctx = SSL_CTX_new(SSLv3_server_method());

	if (NULL == ctx) {
		reason = "SSLErrorNoCtx";
		goto done;
	}

	SSL_CTX_set_mode(ctx, SSL_MODE_AUTO_RETRY);
	SSL_CTX_set_options(ctx, SSL_OP_TLS_ROLLBACK_BUG);

	/* An SSL group will be HIGH, MEDIUM, LOW, or ALL (among other OpenSSL cipher selections). */
	cipher_group = bfconf_get("ssl_cipher_group", NULL);  if (cipher_group == NULL) cipher_group = "ALL";
	/* When FIPS is enabled, change the cipher group to HIGH. */
	#ifdef OPENSSL_FIPS
		if (FIPS_mode()) cipher_group = "HIGH";
	#endif
	/* But allow the cipher_override to be anything, if you set this you should know what you're doing. */
	cipher_override = bfconf_get("ssl_cipher_override", NULL);

	if (cipher_override != NULL) {
		if (1 != SSL_CTX_set_cipher_list(ctx, cipher_override)) {
			reason = "SSLErrorInvalidCipher";
			goto done;
		}
	} else {  /* ALL is the default */
		if (1 != SSL_CTX_set_cipher_list(ctx, cipher_group)) {
			reason = "SSLErrorInvalidCipher";
			goto done;
		}
	}

	/* Set the password for accessing the cert, if necessary */
	if (ssl_cert_password != NULL)
		SSL_CTX_set_default_passwd_cb_userdata(ctx, ssl_cert_password);

	/* Load the key store that contains the cert.  It cannot be NULL at this point. */
	if (1 != SSL_CTX_use_certificate_chain_file(ctx, ssl_cert_location)) {
		reason = "SSLErrorBadCert";
		goto done;
	}

	BFTRACE("Setting key password in default userdata.");
	/* Set the password for accessing the private key */
	if (ssl_key_password != NULL)
		SSL_CTX_set_default_passwd_cb_userdata(ctx, ssl_key_password);
	else
		BFTRACE("Key password is null.");

	BFTRACE("Getting private key from PEM.");
	if (1 != SSL_CTX_use_PrivateKey_file(ctx, ssl_key_location, SSL_FILETYPE_PEM)) {
		reason = "SSLErrorBadPKeyFile";
		goto done;
	}

	BFTRACE("Checking private key from PEM.");
	if (1 != SSL_CTX_check_private_key(ctx)) {
		reason = "SSLErrorBadPKey";
		goto done;
	}

	BFTRACE("Getting CA store information.");
	/* Load the trust store that contains the signers. */
	ssl_ca_location = bfconf_get("ssl_ca_location", NULL);
	ssl_ca_password = getPasswordFromAgentConfig ("ssl_ca_password");
	
	sprintf(buf, "CA location: %s", ssl_cert_location);
	BFTRACE(buf);

	/* Set the password for accessing the CA store, if required. */
	if (ssl_ca_password != NULL)
		SSL_CTX_set_default_passwd_cb_userdata(ctx, ssl_ca_password);

	BFTRACE("Checking the CA store.");
	if (NULL != ssl_ca_location) {
		if (1 != SSL_CTX_load_verify_locations(ctx, ssl_ca_location, NULL)) {
			reason = "SSLErrorBadCA";
			goto done;
		}

		/* Check if client_authentication verification is desired. */
		client_authentication = bfconf_get("ssl_client_authentication", NULL);
		
		if (client_authentication != NULL && !strcmp("true", client_authentication)) {

			/* The list of CA certs sent to the client must be set this way. */
			cert_names = SSL_load_client_CA_file(ssl_ca_location);
			if (cert_names != NULL)
				SSL_CTX_set_client_CA_list(ctx, cert_names);

			SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, verify_SSL_callback);
		} else {
			SSL_CTX_set_verify(ctx, SSL_VERIFY_NONE, NULL);
		}
	}

done:
	if (ssl_ca_location != NULL) free(ssl_ca_location);
	if (ssl_cert_location != NULL) free(ssl_cert_location);
	if (ssl_key_location != NULL) free(ssl_key_location);
	if (cipher_group != NULL) free(cipher_group);
	if (cipher_override != NULL) free(cipher_override);
	if (fips_enabled != NULL) free(fips_enabled);
	if (client_authentication != NULL) free(client_authentication);

	if (NULL != reason) {
		sprintf(buf, "Failure reason: %s", reason);
		BFTRACE(buf);
		sbuf = ERR_error_string(err, err_buf);
		sprintf(buf, "OpenSSL error string: %s", sbuf);
		BFTRACE(buf);
		send_msg("SSL", reason, "s", sbuf);
		if (NULL != ctx) {
			SSL_CTX_free(ctx);
			ctx = NULL;
		}
		return SS_FAILED;
	}
	
	BFTRACE("Returning from init_CTX.");
	return (NULL != ssl_ca_location) ? SS_CERTIFIED : SS_SECURE;
}

void stop_SSL(void) {
	send_goodbye();
	if (ssl) {
		SSL_shutdown(ssl);
		SSL_free(ssl);
		ssl = NULL;
	}

	close(0);
	close(1);
	close(2);

#ifdef WIN32
	if (win_sock) {
		shutdown(ifd, 0);
		closesocket(ifd);
		WSACleanup();
	}
#else /* !WIN32 */
	if (ifd > 2) { close(ifd); }
	if (ofd > 2 && ifd != ofd) { close(ofd); }
#endif /* WIN32/! */
}

void *start_SSL(void) {
	char err_buf[1024];
	const char *err_text;
	const char *sbuf=NULL;
	int err=0, rc=0;
	X509* client_cert;
	
	BFTRACE("In start_SSL");

	/* Specify the context configuration options */
	ssl_state = init_CTX();

	/* You cannot send a hello in the middle of an SSL handshake.  Defer
	   the SSL hello until the end.  Send the TCP hello now, if the state 
	   is SS_OK. 
	 */
	switch (ssl_state) {
		case SS_OK:
			send_hello(CODE_HELLO);
			return NULL;
		default:
			break;  /* Continue to do the SSL handshake. */
	}
	
	BFTRACE("Calling SSL_new");
	/* Create a new SSL context */
	ssl = SSL_new(ctx);
	if (NULL == ssl) {
		sbuf = ERR_error_string(err, err_buf);
		send_msg("SSL", "SSLErrorNoSSL", "s", sbuf);
		ssl_state = SS_FAILED;
		return NULL;
	}

	/* hRemote is the duplicate file descriptor from the dispatch thread. */
	SSL_set_fd(ssl, hRemote);
	SSL_set_accept_state(ssl);
	BFTRACE("Calling SSL_accept.");
	rc = SSL_accept(ssl);
	sprintf(err_buf, "SSL_accept rc=%d (1 is good, 0 is handshake failure, <0 is fatal)", rc);
	BFTRACE(err_buf);

	if (rc <= 0) {
		err = SSL_get_error(ssl, rc);
		sprintf(err_buf, "Error code: %d", err);
		BFTRACE(err_buf);
		ERR_error_string_n(ERR_get_error(), err_buf, sizeof(err_buf)); 
        BFTRACE(err_buf);

		switch (err) 
		{ 
 			case SSL_ERROR_NONE:  
				sprintf(err_buf, "No error reported.");
				BFTRACE(err_buf);
 				break; 
 			case SSL_ERROR_WANT_WRITE: 
				sprintf(err_buf, "Error: SSL_ERROR_WANT_WRITE");
				BFTRACE(err_buf);
 				break; 
 			case SSL_ERROR_WANT_READ: 
				sprintf(err_buf, "Error: SSL_ERROR_WANT_READ");
				BFTRACE(err_buf);
 				break; 
 			case SSL_ERROR_WANT_X509_LOOKUP: 	
				sprintf(err_buf, "Error: SSL_ERROR_WANT_X509_LOOKUP");
				BFTRACE(err_buf);
 				break; 
 			case SSL_ERROR_SYSCALL: {
				int saved_errno = errno;
				err_text = strerror(errno);
				if (NULL == err_text) { err_text = "???"; }
				sprintf(err_buf, "Error: SSL_ERROR_SYSCALL, %d: %s",
						saved_errno, err_text);
				BFTRACE(err_buf);
 				break;
			}
 			case SSL_ERROR_SSL: 
				sprintf(err_buf, "Error: SSL_ERROR_SSL");
				BFTRACE(err_buf);
				break;
 			case SSL_ERROR_ZERO_RETURN: 
				sprintf(err_buf, "Error: SSL_ERROR_ZERO_RETURN");
				BFTRACE(err_buf);
				break;
 		} 
	
		if (sbuf != NULL) send_msg("SSL", "SSLErrorHandshake", "s", sbuf);
		ssl_state = SS_FAILED;
		SSL_free(ssl);
		return NULL;
	}

	sbuf = SSL_get_cipher_name(ssl);

	if (sbuf != NULL) {
		sprintf(err_buf, "Cipher chosen: %s", sbuf);
		BFTRACE(err_buf);
	}

	client_cert = SSL_get_peer_certificate (ssl);
	
	if (client_cert != NULL) {
		sbuf = X509_NAME_oneline (X509_get_subject_name (client_cert), 0, 0);
		sprintf(err_buf, "Client certificate: %s", sbuf);
		BFTRACE(err_buf);
	}

	switch (ssl_state) {
		case SS_OK:
			BFTRACE("ssl_state = SS_OK");
			send_hello(CODE_HELLO);
			return NULL;
		case SS_SECURE:
			BFTRACE("ssl_state = SS_SECURE");
			send_hello(CODE_SECURE_HELLO);
			break;
		case SS_CERTIFIED:
			BFTRACE("ssl_state = SS_CERTIFIED");
			send_hello(CODE_CERTIFIED_HELLO);
			break;
		default:
			return NULL;
	}

	return ssl;
}


void report_SSL(void) {
	char *line;
	X509 *cert;

	if (!ssl) {
		send_msg("SSL", "SSLPlainText", NULL);
		return;
	}

	send_msg("SSL", "SSLStartTLS", "s", SSL_get_cipher_name(ssl));
	if (ssl_state != SS_CERTIFIED) {
		send_msg("SSL", "SSLUntrustedCert", NULL);
	}

	/* Get certificates (if available) */
	cert = SSL_get_peer_certificate(ssl);
	if (NULL != cert) {
		line = X509_NAME_oneline(X509_get_issuer_name(cert), 0, 0);
		if (NULL != line) {
			send_msg("SSL", "SSLPeerCertIssuer", "s", line);
		}
		line = X509_NAME_oneline(X509_get_subject_name(cert), 0, 0);
		if (NULL != line) {
			send_msg("SSL", "SSLPeerCertSubject", "s", line);
		}
		X509_free(cert);
	}
}

#define io_read(buf, len) (ssl ? SSL_read(ssl, buf, len) : platform_protocol_read(buf, len))
#define io_write(buf, len) (ssl ? SSL_write(ssl, buf, len) : platform_protocol_write(buf, len))

#else /* !HAVE_SSL */

/*================================================================
 * Compiled without OpenSSL
 *----------------------------------------------------------------*/

void *start_SSL(void) {
	send_hello(CODE_HELLO);
	ssl_state = SS_OK;
	return NULL;
}

void stop_SSL(void) {
	send_goodbye();
}

void report_SSL(void) {
	/* Nothing to report, for now.  Might want to report that SSL is not
	 * being used.
	 */
}

#define io_read(buf, len) platform_protocol_read(buf, len)
#define io_write(buf, len) platform_protocol_write(buf, len)

#endif /* HAVE_SSL */



#define PUTS_BUF_SIZE (BFBUFSIZ << 2)
static char *puts_buf = NULL;


int io_puts(const char *s) {
	int len = strlen(s);

	/* Writing UTF-8 to a socket ... */
	if (platform_protocol_has_socket() ||
			BFCONF_EXISTS(disable_transcode))
	{
#ifdef ZOS
		/* On z/OS, our UTF-8 data is transcoded to EBCDIC, and we
		 * need to trancode it back to real UTF-8 before we send it.
		 */
		int ret;
		char *p = malloc(len);
		memcpy(p, s, len);
		convert_etoa_l(p, len);
		ret = io_write(p, len);
		free(p);
		return ret;
#else /* !ZOS */
		/* On other platforms, UTF-8 is just UTF-8. */
		return io_write(s, len);
#endif /* ZOS */
	}

	/* Writing to the OS, so use its console encoding ... */
	if (platform_utf8_is_console()) {
		return io_write(s, len);
	}
	if (NULL == puts_buf) {
		puts_buf = malloc(PUTS_BUF_SIZE);
	}
	len = convert_utf8_to_console(s, len, puts_buf, PUTS_BUF_SIZE);
	return io_write(puts_buf, len);
}

#define TELNET_SE   (240)
#define TELNET_NOP  (241)
#define TELNET_MARK (242)
#define TELNET_BRK  (243)
#define TELNET_IP   (244)
#define TELNET_AO   (245)
#define TELNET_AYT  (246)
#define TELNET_EC   (247)
#define TELNET_EL   (248)
#define TELNET_GA   (249)
#define TELNET_SB   (250)
#define TELNET_WILL (251)
#define TELNET_WONT (252)
#define TELNET_DO   (253)
#define TELNET_DONT (254)
#define TELNET_IAC  (255)

#define TNO_BINARY      (0)
#define TNO_SUPPRESS_GA (3)
#define TNO_TIMING_MARK (6)

#define GETS_UTF8_SIZE (BFBUFSIZ << 1)
static char gets_utf8[GETS_UTF8_SIZE];
static int gets_utf8_len = 0;

typedef enum {
	TNS_RESET,
	TNS_IAC,
	TNS_WILL,
	TNS_WONT,
	TNS_DO,
	TNS_DONT
} telnet_state_t;

static telnet_state_t telnet_state = TNS_RESET;
static const uint8_t TELNET_RESPONSE_AYT[] = {
	/* "\r\n[IBM Rational Build Forge Agent]\r\n" in ASCII */
	0x0D, 0x0A, 0x5B,
	0x49, 0x42, 0x4D, 0x20,
	0x52, 0x61, 0x74, 0x69, 0x6F, 0x6E, 0x61, 0x6C, 0x20,
	0x42, 0x75, 0x69, 0x6C, 0x64, 0x20,
	0x46, 0x6F, 0x72, 0x67, 0x65, 0x20,
	0x41, 0x67, 0x65, 0x6E, 0x74,
	0x5D, 0x0D, 0x0A
};

int io_ayt(void) {
	return io_write(TELNET_RESPONSE_AYT, sizeof(TELNET_RESPONSE_AYT));
}

static void io_gets_telnet_option(uint8_t code) {
	uint8_t buf[3];
	buf[0] = TELNET_IAC;
	/* buf[1] = response */
	buf[2] = code;

	switch (code) {
		/* Options we will claim to support, even though we don't actually
		 * do anything special for them.
		 */
		case TNO_BINARY:
		case TNO_SUPPRESS_GA:
		case TNO_TIMING_MARK:
		{
			switch (telnet_state) {
				case TNS_WILL: buf[1] = TELNET_DO;   break;
				case TNS_WONT: buf[1] = TELNET_DONT; break;
				case TNS_DO:   buf[1] = TELNET_WILL; break;
				case TNS_DONT: buf[1] = TELNET_WONT; break;
				default: /* NOT REACHED */ return;  
			}
			break;
		}

		/* We don't even understand what they're asking us to do,
		 * so refuse to cooperate.
		 */
		default:
		{
			switch (telnet_state) {
				case TNS_WILL: /* FALL THROUGH */
				case TNS_WONT: buf[1] = TELNET_DONT; break;
				case TNS_DO:   /* FALL THROUGH */
				case TNS_DONT: buf[1] = TELNET_WONT; break;
				default: /* NOT REACHED */ return;
			}
		}
	}

	io_write(buf, 3);
}

/* At some point, we discovered a 0xFF byte in the protocol stream.  That
 * is not a valid byte in UTF-8, so we will assume that a user is connected
 * to the agent using telnet and that it came from the telnet client.  This
 * is annoying, but we'll deal with it as best as we can by responding
 * appropriately when we see telnet control sequences.  If we encounter an
 * incomplete sequence, we'll save it for later.  Some sequences can be
 * quite long, so we give up if it looks like a sequence is too long to
 * bother trying to honor it.  In this case, we discard the IAC and
 * release the rest of the data to be processed as (alleged) UTF-8.
 */
static void io_gets_telnet(char *telnet_mark) {
	char *p = telnet_mark;
	char *pend = &gets_utf8[gets_utf8_len];

	/* Miniature FSM for telnet option negotiation */
	while (1) {
		/* Exhausted the buffer.  Truncate it at the telnet sequence. */
		if (p >= pend) {
			gets_utf8_len = telnet_mark - gets_utf8;
			return;
		}

		switch (telnet_state) {
			case TNS_RESET:
			{
				p = memchr(telnet_mark, TELNET_IAC, pend - telnet_mark);
				if (NULL == p)
					return;
				++p;
				telnet_state = TNS_IAC;
				continue;
			}

			case TNS_IAC:
			{
				switch ((uint8_t)(*p++)) {
					case TELNET_AYT:
						io_ayt();
						goto completed;
					case TELNET_EC:
						if (telnet_mark > gets_utf8)
							--telnet_mark;
						goto completed;
					case TELNET_EL:
						telnet_mark = memrchr(gets_utf8, 0x0A, gets_utf8_len);
						if (NULL == telnet_mark)
							telnet_mark = gets_utf8;
						goto completed;
					case TELNET_WILL:
						telnet_state = TNS_WILL;
						continue;
					case TELNET_WONT:
						telnet_state = TNS_WONT;
						continue;
					case TELNET_DO:
						telnet_state = TNS_DO;
						continue;
					case TELNET_DONT:
						telnet_state = TNS_DONT;
						continue;
					case TELNET_IAC:
						++telnet_mark;
						goto completed;
					default:
						goto completed;
				}
			}

			default: /* WILL WONT DO DONT */
				io_gets_telnet_option((uint8_t)*p++);
				goto completed;
		}

completed:
		gets_utf8_len -= p - telnet_mark;
		pend = &gets_utf8[gets_utf8_len];
		telnet_state = TNS_RESET;
		if (telnet_mark < &gets_utf8[gets_utf8_len]) {
			memmove(telnet_mark, p, pend - telnet_mark);
			p = telnet_mark = memchr(telnet_mark, TELNET_IAC, pend - telnet_mark);
			if (NULL != telnet_mark)
				continue;
		}
		return;
	}
}

/* Read UTF-8 data directly into the buffer ... */
static int io_gets_fill_utf8(void) { while (1) {
	int len;
	char *p = &gets_utf8[gets_utf8_len];
	char *q = p;

	len = io_read(p, GETS_UTF8_SIZE - gets_utf8_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_utf8: io_read -> len=%d",
				len);
	}

	if (len <= 0) { return len; }

	p = &gets_utf8[gets_utf8_len];
	gets_utf8_len += len;

	/* If this is a person using a telnet, then we might see telnet
	 * control sequences.  This deals with them if there are any.
	 */
	if (!BFCONF_EXISTS(disable_telnet_support)) {
		if (telnet_state != TNS_RESET || NULL != (q = memchr(p, 0xFF, len))) {
			io_gets_telnet(q);
			len = &gets_utf8[gets_utf8_len] - p;
			if (len <= 0)
				continue;
		}
	}

#ifdef ZOS
	convert_atoe_l(p, len);
#endif
	return gets_utf8_len;
}}


#define GETS_CONSOLE_SIZE (BFBUFSIZ)
/* Read from the OS, using its console encoding ... */
static char *gets_console = NULL;
static int gets_console_len = 0;

/* Reads from the console until there is a suitable boundary for
 * transcoding.  If the buffer fills completely and there is no
 * candidate boundary, then return the entire buffer.
 */
static int io_gets_fill_console(void) { while (1) {
	int len = io_read(&gets_console[gets_console_len],
			GETS_CONSOLE_SIZE - gets_console_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_console: io_read -> len=%d",
				len);
	}

	/* The read returned an error or EOF.  If we have data buffered,
	 * then we flush it.  Otherwise, pass on the error/EOF result.
	 */
	if (len <= 0)
		return (gets_console_len > 0) ? gets_console_len : len;
	gets_console_len += len;

	/* See if we can find a good stopping point for transcoding. */
	len = convert_console_find_boundary(gets_console, gets_console_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_console: "
				"convert_console_find_boundary: %d -> %d",
				gets_console_len, len);
	}
	if (len > 0)
		return len;

	/* If the buffer is full, give up and transcode the whole thing. */
	if (gets_console_len >= GETS_CONSOLE_SIZE) {
		if (NULL != activity_file) {
			BFTRACE("IO io_gets_fill: console buffer overflow");
		}
		return GETS_CONSOLE_SIZE;
	}

	/* Not at a stopping point and still have room, so keep going... */
}}




static int io_gets_fill_transcode(void) {
	int len, converted;

	len = io_gets_fill_console();
	if (len <= 0) {
		if (gets_console_len <= 0)
			return len;
		len = gets_console_len;
	}

	/* Convert a block of console text to UTF-8. */
	converted = convert_console_to_utf8(gets_console, len,
			&gets_utf8[gets_utf8_len], GETS_UTF8_SIZE - gets_utf8_len);
	if (NULL != activity_file) {
		activity_write(BFHERE
				"IO io_gets_fill_transcode: "
				"console [%d - %d] -> UTF8 [%d + %d]",
				gets_console_len, len,
				gets_utf8_len, converted);
	}

	gets_console_len -= len;
	memmove(gets_console, &gets_console[len], gets_console_len);
	return gets_utf8_len += converted;
}

static int (*io_gets_fill_fn)(void) = NULL;
#define io_gets_fill() ((*io_gets_fill_fn)())

char *io_gets(char *buf, int maxlen) {
	int i=0;

	/* figure out if we are going to fill the buffer using a
	 * network socket or the console.
	 */
	if (NULL == io_gets_fill_fn) {
		if (platform_protocol_has_socket() ||
				BFCONF_EXISTS(disable_transcode)) {
			io_gets_fill_fn = &io_gets_fill_utf8;
		} else {
			io_gets_fill_fn = &io_gets_fill_transcode;
			gets_console = malloc(GETS_CONSOLE_SIZE);
		}
	}

	/* If we fill the buffer, then stop regardless of whether or not
	 * we find a newline.
	 */
	while (i < maxlen-1) {
		/* If we need more chars than are currently in the UTF-8 buffer,
		 * then we need to fill it.  If we can't, then return what we
		 * have, or NULL if we don't have anything.
		 */
		while (i >= gets_utf8_len) {
			if (io_gets_fill() <= 0) {
				if (gets_utf8_len <= 0) {
					if (NULL != activity_file) {
						BFTRACE("IO io_gets: NULL");
					}
					return NULL;
				}
				break;
			}
		}

		/* Non-control character -- preserve and continue */
		if (gets_utf8[i] & ~0x1F) {
			++i;
			continue;
		}

		/* Tolerate tab chars */
		if (gets_utf8[i] == '\t') {
			++i;
			continue;
		}

		/* Map CR to LF */
		if (gets_utf8[i] == '\r') { gets_utf8[i] = '\n'; }

		/* If it's \n, we have a full line. */
		if (gets_utf8[i] == '\n') {
			++i;
			break;
		}

		/* Backspace?  You must be using the broken Windows telnet...  We
		 * only honor it if there is something to delete; otherwise, we
		 * just discard it.
		 */
		if (gets_utf8[i] == '\b') {
			if (i > 0) {
				--i;
				gets_utf8_len -= 2;
				memmove(&gets_utf8[i], &gets_utf8[i+2], gets_utf8_len - i);
			} else {
				--gets_utf8_len;
				memmove(&gets_utf8[i], &gets_utf8[i+1], gets_utf8_len - i);
			}
			continue;
		}

		/* Any other control char -- immediately hang up on the client */
		io_ayt();
		return NULL;
	}

	/* Return the data to the caller and remove it from the buffer. */
	memcpy(buf, gets_utf8, i);
	buf[i] = 0;
	gets_utf8_len -= i;
	memmove(gets_utf8, &gets_utf8[i], gets_utf8_len);

	/* Don't trust the client to speak UTF-8 properly... Some things, like
	 * overlong encoding, can be dangerous if they are allowed.
	 */
	convert_utf8_clean(buf, i, 1);

	if (NULL != activity_file) {
		activity_write(BFHERE "IO io_gets: [%d]=\"%s\"", i, buf);
	}
	return buf;
}


