/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"
#include "md5.h"

int BFGetPath(const char *path, int(*)(const char *, const struct stat *, int));

#ifdef WIN32
#  define FTW_F     1
#  define FTW_D     2
#  define FTW_DNR   3
#  define FTW_SL    4
#  define FTW_NS    5
# else /* !WIN32 */
#  if HAVE_FTW_H
#    undef HAVE_FTS_H
#    include <ftw.h>
#  endif /* HAVE_FTW_H */
#  ifdef HAVE_FTS_H
#    include <fts.h>
#    define FTW_F       FTS_F
#    define FTW_D       FTS_D
#    define FTW_SL      FTS_SL
#    define FTW_NS      FTS_NS
#    ifdef FTS_DNR
#      define FTW_DNR   FTS_DNR
#    endif /* FTS_DNR */
#  endif /* HAVE_FTS_H */
#endif /* WIN32/! */


/*
 * CRC-32-MPEG2 (aka IEEE 802.3 / V.42)
 */
static uint32_t *crc_table = NULL;
static void crc_gen_table(void)			/* build the crc table */
{
	uint32_t crc, poly = 0xEDB88320L;
	int	i, j;

	crc_table = calloc(256, sizeof(uint32_t));

	for (i = 0; i < 256; i++) {
		crc = i;
		for (j = 8; j > 0; j--) {
			if (crc & 1)
				crc = (crc >> 1) ^ poly;
			else
				crc >>= 1;
		}
		crc_table[i] = crc;
	}
}

static uint32_t crc_string(const char *s)	/* calculate the crc value */
{
	register uint32_t crc = 0xFFFFFFFF;
	if (NULL != s) {
		while ('\0' != *s) {
			crc = (crc >> 8) ^ crc_table[(crc ^ *s++) & 0xFF];
		}
	}
	return crc ^ 0xFFFFFFFF;
}


/* CRF -- This has never been used...
static uint32_t crc_file(char *path)
{
	register uint32_t crc = 0xFFFFFFFF;
	register char *p;
	int fd;
	size_t	bytes;
	char	*buf;

	fd = platform_open2(path, O_RDONLY|O_BINARY);
	if (fd == -1) return 0;

	buf = malloc(BFBUFSIZ);
	while ((bytes = read(fd, buf, BFBUFSIZ-1)) > 0) {
		buf[bytes] = 0;
		for (p = buf; *p; p++)
			crc = (crc>>8) ^ crc_table[(crc ^ *p++) & 0xFF ];
	}
	close(fd);
	free(buf);
	return crc ^ 0xFFFFFFFF;
}
*/


/*
 * Callback functions
 */

/* callback function to transmit a file/entity */
static char *cb_read_buf = NULL;
static int cb_read(const char *p, const struct stat *sb, int flags)
{
	int     err;
	int     fd;
	size_t  bytes;
	off_t   offset, blocks;
	char *  path = CleanPath(strdup(p));

	err = 0;
	switch (flags) {
		case FTW_F:
			blocks = (sb->st_size + BFXFERBLOCKSIZE - 1) / BFXFERBLOCKSIZE;
			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd != -1) {
				offset = (off_t)0;
				while ((bytes = read(fd, cb_read_buf, BFXFERBLOCKSIZE)) > 0) {
					send_file_info(
							(off_t)offset,
							(long)bytes,
							(int)sb->st_mode,
							(long)sb->st_size,
							(long)sb->st_mtime,
							path);
					send_file_data((const uint8_t *)cb_read_buf, bytes);
					offset += bytes;
				}
				close(fd);
			}
			break;
		case FTW_D:
#ifdef FTW_DNR
		case FTW_DNR:
#endif
			send_dir((int)sb->st_mode, (long)sb->st_mtime, path);
			break;
		case FTW_SL:
			break;
	}

	free(path);
	return 0;
}

/* callback function to transmit a checkpoint for a file/entity */
static int cb_checkpoint(const char *p, const struct stat *sb, int flags)
{
	char *          info;
	char *          buf;
	uint32_t        crc;
	md5_state_t     ctx;
	md5_byte_t      digest[16];
	int             fd;
	int             got;
	off_t           block;
	off_t           blocks;
	char *          path = CleanPath(strdup(p));

	info = calloc(1, BFBUFSIZ);
	crc = crc_string(path);
	switch (flags) {
		case FTW_F:
			block = 0;
			blocks = (sb->st_size + BFXFERBLOCKSIZE - 1) / BFXFERBLOCKSIZE;
			fd = platform_open2(path, O_RDONLY|O_BINARY);
			if (fd != -1) {
				md5_init(&ctx);
				buf = calloc(1, BFBUFSIZ);

				while ((got = read(fd, buf, BFBUFSIZ)) > 0) {
					md5_append(&ctx, (const md5_byte_t *)buf, (size_t)got);
					if ((++block & 65535) == 0) {
						activity_write(BFHERE "cb_checkpoint: md5 block %llu of %llu",
								(unsigned long long)block,
								(unsigned long long)blocks );
					}
				}
				close(fd);
				md5_finish(&ctx, digest);
				for (fd = 0; fd < 16; fd++) {
					sprintf(buf + (fd << 1), "%02x", digest[fd]);
				}
				send_checkpoint("FILE", crc, (long)sb->st_mtime, digest, path);
				free(buf);
			} else {
				free(path);
				return 0;
			}
			break;
#ifdef FTW_DNR
		case FTW_DNR:
#endif
		case FTW_D:
			send_checkpoint("DIR", crc, (long)sb->st_mtime, NULL, path);
			break;
		case FTW_SL:
			send_checkpoint("SYMLINK", crc, (long)sb->st_mtime, NULL, path);
			break;
		default:
			break;
	}
	free(info);
	free(path);
	return 0;
}



/*
 * Implementation for "cmd read".
 *
 * Read a local file/tree and send it to stdout.
 */
int FuncRead(void)
{
	char	*path, *recurse, *checkpoint, *p, *userdata;
	char	*timeout;
	int 	err, rc = 0;
	int 	mapfail;
	size_t	userdata_size;

	/*  locate the file/dir element */
	if (NULL == (path = request_get(TAG_PATH))) {
		send_missing(CMD_READ, TAG_PATH);
		return 1;
	}
	path = strdup(path); /* this will get free'd later */

	/*  locate the userdata element */
	userdata = request_get_binary(TAG_USERDATA, &userdata_size);

	/*  locate the client timeout element */
	timeout = request_get(TAG_TIMEOUT);
	if (NULL == timeout) { timeout = "0"; }

	/*  optional recursive flag */
	recurse = request_get(TAG_RECURSIVE);

	/*  optional checkpoint flag, which implies recursive */
	checkpoint = request_get(TAG_CHECKPOINT);
	if (checkpoint && !recurse) { recurse = checkpoint; }

	StoreUserData(userdata, userdata_size);

	mapfail = BFCONF_EXISTS(map_drive_is_failure);
	if ( (platform_map_drives(defMapStr) != 0 && mapfail) ||
			(platform_map_drives(mapStr) != 0 && mapfail) )
	{
		send_msg("MAP", "MapFail", NULL);
		FreeUserData();
		platform_unmap_drives(0);
		return 2;
	}

	ClobberPasswords();

	if (clearcase_view) {
		size_t len = strlen(this_view()) + strlen(clearcase_view) + strlen(path) + 3;
		char *ccpath = calloc(1, len);
		format_message(ccpath, "%s/%s/%s", this_view(), clearcase_view, path);
		free(path);
		path = ccpath;
	}

	/* process the clearcase view/vob start/mount functions */
	CCStartViews(atoi(timeout));
	CCMountVobs(atoi(timeout));

	/*  assemble and clean the final pathname. clean the
	 *  pieces so we can use them individually later.
	 */
	CleanPath(path);
	path = env_expand_var((p = path), 0);
	free(p);	/* Free the previous value of path */

	/* if not recursive, path must point to a file */
	if (!checkpoint && !recurse && !isFile(path)) {
		send_msg("READ", "ReadNotFile", "s", path);
		rc = 3;
	} else if (0 != platform_access(path, 0)) {
		/* We can't checkpoint or read something that isn't there. */
		send_msg("READ", "ReadNotFound", "s", path);
		rc = 4;
	} else {
		if (checkpoint) {
			if (NULL == crc_table) { crc_gen_table(); }
			err = BFGetPath(path, cb_checkpoint);
		} else {
			cb_read_buf = malloc(BFXFERBLOCKSIZE);
			err = BFGetPath(path, cb_read);
			free(cb_read_buf);
			cb_read_buf = NULL;
		}

		if (err) {
			send_msg("READ", "ReadFail", "is", err, path);
			rc = 5;
		} else {
			send_msg("READ", "ReadOk", "s", path);
		}
	}

	free(path);
	platform_unmap_drives(0);
	FreeUserData();
	return rc;
}

