/*
Copyright (C)2007-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/


#include "bf_agent.h"

#define INITIAL_HASH_SIZE (11)

#ifdef REQUEST_DEBUG
#define D(x) x
#else
#define D(x) /**/
#endif

static hash_ctx *ctx = NULL;
static hash_table *request = NULL;

struct _TagData {
	char	*buf;				/* malloc'd data buffer */
	size_t	size;				/* size of malloc'd buffer */
	int     stream_fd;
};
typedef struct _TagData TagData;

static void release_tag_data(void *d) {
	TagData *data = d;
	if (data != NULL) {
		if (data->stream_fd >= 0) {
			close(data->stream_fd);
			if (data->buf)
				platform_unlink(data->buf);
		}
		if (data->buf)
			free(data->buf);
		free(data);
	}
}

void request_new(void) {
	D(printf("request_new\n"));
	if (request == NULL) {
		if (ctx == NULL) {
			ctx = hash_ctx_new();
			ctx->free_value_fn = release_tag_data;
		}
		request = hash_new(ctx, INITIAL_HASH_SIZE);
	} else {
		request_del_creds();
		hash_clear(request);
		hash_resize(request, INITIAL_HASH_SIZE);
	}
}

void request_destroy(void) {
	D(printf("request_destroy\n"));
	if (request != NULL) {
		request_del_creds();
		hash_destroy(request);
		request = NULL;
	}
}

int request_size(void) {
	D(printf("request_size -> %d\n", request->count));
	return request->count;
}


void request_dump(void) {
	const char *key;
	char *buf = malloc(BFBUFSIZ);
	char *p;
	TagData *data;
	size_t len, left;

	BFTRACE("request_dump:");
	if (request != NULL) {
		hash_iter *hi = hash_iter_new(request);
		while (hash_iter_next(hi)) {
			key = hi->entry->key;
			data = hi->entry->value;

			if (strcmp(key, TAG_PASSWORD) == 0 || strcmp(key, TAG_ENCPASS) == 0) {
				activity_write(BFHERE "[%s] => *****", key);
				continue;
			}

			len = format_message(buf, "[%s] => %u: [", key, (unsigned int)data->size);
			p = &buf[len];

			if (strcmp(key, TAG_USERDATA) == 0) {
				if (data->size == 0) {
					*p++ = ']';
					*p = '\0';
				} else {
					size_t bytes = data->size;
					const char *suffix = "]";
					char *hex;

					if (bytes > (BFBUFSIZ >> 2)) {
						bytes = BFBUFSIZ >> 2;
						suffix = " ... ]";
					}

					hex = malloc((bytes << 1) + 8);
					hex_encode(hex, (const uint8_t *)data->buf, bytes);
					format_message(p, "%s%s", hex, suffix);
					free(hex);
				}
				BFTRACE(buf);
				continue;
			}

			left = BFBUFSIZ - len - 16;	/* Reserve a bit */
			if (left >= data->size) {
				left = data->size - 1;
				memcpy(p, data->buf, left);
				strcpy(&p[left], "]");
			} else {
				memcpy(p, data->buf, left);
				strcpy(&p[left], " ... ]");
			}
			BFTRACE(buf);
		}
		hash_iter_destroy(hi);
	}
	free(buf);
}

void request_put(const char *tag, char *value) {
	TagData *data = hash_get_str(request, tag);
	int size = 0;
	char *p;

	D(printf("request_put: tag=[%s] value=[%s]\n", tag, value));
	if (data == NULL) {
		data = calloc(1, sizeof(*data));
		data->stream_fd = -1;
		hash_put_str(request, strdup(tag), data);
	} else if (data->stream_fd >= 0) {
		request_put_stream(tag, value, strlen(value));
		return;
	} else {
		size = data->size;
	}
	if (size == 0) {
		free(data->buf);
		data->buf = strdup(value);
		data->size = strlen(value) + 1;
	} else {
		data->buf = realloc(data->buf, size + strlen(value) + 2);
		p = &data->buf[size-1];
		sprintf(p, "\n%s", value);
		data->size += strlen(p);
	}
}

void request_put_binary(const char *tag, void *value, size_t len) {
	TagData *data = hash_get_str(request, tag);
	int size = 0;

	D(printf("request_put_binary: tag=[%s] len=%d\n", tag, (int)len));
	if (data == NULL) {
		data = calloc(1, sizeof(*data));
		data->stream_fd = -1;
		hash_put_str(request, strdup(tag), data);
	} else if (data->stream_fd >= 0) {
		request_put_stream(tag, value, strlen(value));
		return;
	} else {
		size = data->size;
	}
	if (size == 0) {
		if (data->buf != NULL)
			free(data->buf);
		data->size = len;
		data->buf = malloc(len);
		memcpy(data->buf, value, len);
	} else {
		data->size += len;
		data->buf = realloc(data->buf, size + len);
		memcpy(data->buf + size, value, len);
	}
}

void request_put_stream(const char *tag, void *value, size_t len) {
	TagData *data = hash_get_str(request, tag);
	char *p = value;

	D(printf("request_put_stream: tag=[%s] len=%d\n", tag, (int)len));
	if (data == NULL) {
		data = calloc(1, sizeof(*data));
		hash_put_str(request, strdup(tag), data);
		data->stream_fd = platform_tempfd(&data->buf);
		data->size = strlen(data->buf) + 1;
	} else if (data->stream_fd < 0) {
		request_put_binary(tag, value, strlen(value));
		return;
	}
	while (len > 0) {
		int bytes = write(data->stream_fd, p, len);
		if (bytes < 0) {
			/* Retry once; maybe we'll get lucky... */
			bytes = write(data->stream_fd, p, len);
			if (bytes < 0)
				break;
		}
		p += bytes;
		len -= bytes;
	}
}

int request_del(const char *tag) {
	D(printf("request_del: tag=[%s]\n", tag));
	return hash_del_str(request, tag);
}

static const char *cred_tags[] = {
	TAG_ENCPASS,
	TAG_PASSWORD,
	TAG_USERNAME,
	NULL
};

void request_del_creds(void) {
	const char **tag;
	TagData *data;
	
	D(printf("request_del_creds\n"));
	for (tag = cred_tags; *tag; ++tag) {
		if (NULL != (data = hash_get_str(request, *tag))) {
			if (data->size > 0) {
				memset(data->buf, 0, data->size);
				data->size = 0;
			}
			hash_del_str(request, *tag);
		}
	}
}

int request_has(const char *tag) {
	D(printf("request_has: tag=[%s] -> %d\n", tag, hash_contains_key_str(request, tag)));
	return hash_contains_key_str(request, tag);
}

char *request_get(const char *tag) {
	TagData *data = hash_get_str(request, tag);
	D(printf("request_get: tag=[%s] -> data=[%s]\n", tag, data == NULL ? NULL : data->buf));
	return data == NULL ? NULL : data->buf;
}

void *request_get_binary(const char *tag, size_t *size) {
	TagData *data = hash_get_str(request, tag);
	if (data == NULL) {
		D(printf("request_get_binary: tag=[%s] -> size=0\n", tag));
		*size = 0;
		return NULL;
	}
	*size = data->size;
	D(printf("request_get_binary: tag=[%s] -> size=%d\n", tag, (int)*size));
	return data->buf;
}

int request_get_stream(const char *tag) {
	TagData *data = hash_get_str(request, tag);
	return (NULL == data) ? (-1) : data->stream_fd;
}

/* Note: Don't use a 0x0D in here, as io_puts eats them. */
static const char MAGIC[] = { 'I', 'R', 'B', 'A', 0x20, 0x0A, 0x0A, 0x1A };

void *request_serialize(size_t *size) {
	char *buf, *p, *key;
	size_t need, capacity, keylen;
	hash_iter *hi;
	TagData *data;

	if (NULL == request || 0 == request->count) {
		buf = malloc(sizeof(MAGIC) + 1);
		memcpy(buf, MAGIC, sizeof(MAGIC));
		buf[sizeof(MAGIC)] = 0;
		if (NULL != size)
			*size = sizeof(MAGIC) + 1;
		return buf;
	}

	buf = malloc(capacity = BFBUFSIZ);
	memcpy(buf, MAGIC, sizeof(MAGIC));
	p = &buf[sizeof(MAGIC)];

	hi = hash_iter_new(request);

	while (hash_iter_next(hi)) {
		key = hi->entry->key;
		keylen = hi->entry->keylen;
		data = hi->entry->value;

		need = (p - buf) + 1 + keylen + 4 + data->size;
		if (need > capacity) {
			do { capacity += BFBUFSIZ; } while (need > capacity);
			buf = realloc(buf, capacity);
		}

		*p++ = (unsigned char)keylen;
		memcpy(p, key, keylen);
		p = &p[keylen];

		*p++ = (char)((data->size >> 24) & 0x7F);
		*p++ = (char)((data->size >> 16) & 0xFF);
		*p++ = (char)((data->size >> 8) & 0xFF);
		*p++ = (char)(data->size & 0xFF);
		memcpy(p, data->buf, data->size);
		p = &p[data->size];
	}

	need = p - buf + 1;
	if (need != capacity)
		buf = realloc(buf, capacity = need);
	*p++ = 0;
	if (NULL != size)
		*size = capacity;
	return buf;
}

int request_deserialize(const void *buf) {
	const unsigned char *p = buf;
	char *key;
	void *value;
	size_t len;

	request_new();
	if (memcmp(buf, MAGIC, sizeof(MAGIC)) != 0)
		return 0;
	p += sizeof(MAGIC);

	while (*p != '\0') {
		len = *p++;
		key = malloc(len);
		memcpy(key, p, len);
		p += len;

		len = *p++ << 24;
		len += *p++ << 16;
		len += *p++ << 8;
		len += *p++;
		value = malloc(len);
		memcpy(value, p, len);
		p += len;

		request_put_binary(key, value, len);
	}
	return 1;
}

