/*
Copyright (C)2003-2008, IBM, Inc. All rights reserved. 
IBM, Inc. owns the copyright and other intellectual 
property rights in this software. Duplication or use of the 
Software is not permitted except as expressly provided in a 
written agreement between your company and IBM, Inc.
*/

#include "bf_agent.h"


static char *datablock_cache = NULL;
static size_t datablock_cache_len = 0;
static size_t datablock_cache_size = 0;

static char *output_buffer = NULL;
static size_t output_buffer_size = 0;

int did_hello = 0;

extern int heartbeat_interval;

/*
 * Flushes the output buffer -- either to the datablock cache or
 * directly to io_puts()...
 */
static int flush_output(void) {
	size_t len;

	if (NULL == output_buffer) { return 0; }

	/* If we are caching... */;
	if (NULL != datablock_cache) {
		len = strlen(output_buffer) + 1;

		/* If it would overflow the cache, then flush to make more room. */
		if (datablock_cache_len + len >= datablock_cache_size) {
			if (!send_cache_flush()) { return 0; }
		}

		/* If it fits in the cache, then cache it */
		if (len < datablock_cache_size) {
			memcpy(&datablock_cache[datablock_cache_len], output_buffer, len);
			datablock_cache_len += len - 1;
			*output_buffer = '\0';
			return 1;
		}
	}

	/* Not caching or output_buffer is too big to fit */
	len = io_puts(output_buffer);
	*output_buffer = '\0';
	return (int)len;
}


int send_cache_flush(void) {
	int ok = 1;
	if (NULL != datablock_cache && '\0' != *datablock_cache) {
		ok = (io_puts(datablock_cache) > 0) ? 1 : 0;
		*datablock_cache = '\0';
		datablock_cache_len = 0;
	}
	return ok;
}


int send_cache_init(int size) {
	if (size <= 0) { return 0; }

	BUFFER_ENSURE_CAPACITY(
			&datablock_cache,
			&datablock_cache_size,
			size );

out_of_memory:
	return (int)datablock_cache_size;
}


int send_checkpoint(
		const char *type,
		uint32_t crc32,
		long mtime,
		const uint8_t *md5,
		const char *path)
{
	char crc32hex[16];
	char *md5hex = NULL;
	size_t len;

	sprintf(crc32hex, "%x", (unsigned int)crc32);
	if (NULL != md5) {
		md5hex = malloc(48);
		hex_encode(md5hex, md5, 16);
	}

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(type) + (strlen(path) << 2) + 128 );
	len = format_message(output_buffer, "330 %s %08x %ld %s ",
			type,
			(unsigned int)crc32,
			mtime,
			(NULL != md5hex) ? md5hex : "." );
	if (NULL != md5hex) { free(md5hex); }

	buffer_append_escaped(
			&output_buffer,
			&output_buffer_size,
			&len,
			path);
	output_buffer[len++] = '\n';
	output_buffer[len] = '\0';

out_of_memory:
	return flush_output();
}


int send_cont(const char *msg) {
	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(msg) + 32 );
	format_message(output_buffer, "311 CONT %s\n", msg);
	log_write(output_buffer);

out_of_memory:
	return flush_output();
}


int send_debug(const char *message) {
	if (NULL == bf_debug || NULL == message || !did_hello) {
		return 1;
	}
	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			strlen(message) + 16 );
	format_message(output_buffer, "%d DEBUG %s\n",
			CODE_DEBUG, message);

out_of_memory:
	return flush_output();
}



int send_dir(int mode, long mtime, const char *path)
{ 
	size_t idx;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(strlen(path) << 2) + 128 );
	idx = format_message(output_buffer, "340 DIR %o %ld ", mode, mtime);
	buffer_append_escaped(&output_buffer, &output_buffer_size, &idx, path);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output();
}


int send_file_data(const uint8_t *bytes, size_t len)
{
	size_t idx = 4;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(len << 1) + 8 );
	strcpy(output_buffer, "342 ");
	idx += hex_encode(&output_buffer[idx], bytes, len);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output();
}


int send_file_info(
		off_t offset,
		long bytes,
		int mode,
		long size,
		long mtime,
		const char *path)
{
	size_t idx;

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			(strlen(path) << 2) + 128 );
	idx = format_message(output_buffer, "341 FILE %llu %lu %o %ld %ld ",
			(unsigned long long)offset, bytes, mode, size, mtime);
	buffer_append_escaped(&output_buffer, &output_buffer_size, &idx, path);
	output_buffer[idx++] = '\n';
	output_buffer[idx] = '\0';

out_of_memory:
	return flush_output();
}


int send_goodbye(void) {
	char buf[128];
	format_message(buf, "%d %s\n", CODE_GOODBYE, CODE_GOODBYE_MSG);
	return io_puts(buf);
}


static time_t beatmark = 0;
static char beatbuf[32];
int send_heartbeat(void) {
	time_t now;

	time(&now);
	if (now - beatmark > heartbeat_interval) {
		if (!send_cache_flush()) { return 0; }
		beatmark = now;
		format_message(beatbuf, "%d HEARTBEAT %d\n",
				CODE_HEARTBEAT, bf_LoadValue());
		return (io_puts(beatbuf) > 0) ? 1 : 0;
	}
	return 1;
}


int send_hello(int code) {
	const char *msg;
	char buf[128];

	switch (code) {
		case CODE_HELLO:            msg = CODE_HELLO_MSG;            break;
		case CODE_SECURE_HELLO:     msg = CODE_SECURE_HELLO_MSG;     break;
		case CODE_CERTIFIED_HELLO:  msg = CODE_CERTIFIED_HELLO_MSG;  break;
		default: msg = "???";
	}
	did_hello = 1;
	format_message(buf, "%d %s\n", code, msg);
	return io_puts(buf);
}


int send_log(const char *bucket, const char *msg) {
	size_t bucketlen;
	size_t msglen;
	char *p;

	if (NULL == bucket || '\0' == *bucket) {
		bucket = "*";
		bucketlen = 1;
	} else {
		bucketlen = strlen(bucket);
	}

	while (NULL != ( p = strpbrk(msg, "\r\n") )) {
		msglen = (size_t)(p - msg);
		BUFFER_ENSURE_CAPACITY(
				&output_buffer,
				&output_buffer_size,
				bucketlen + msglen + 32);

		p = output_buffer;
		p += format_message(p, "%d %s ", CODE_LOG, bucket);
		memcpy(p, msg, msglen);
		p += msglen;
		*p++ = '\n';
		*p = '\0';
		log_write(output_buffer);
		if (!flush_output()) { return 0; }

		msg = &msg[msglen];
		if ('\0' == *msg) { break; }
		msg = ('\r' == *msg && '\n' == msg[1]) ? &msg[2] : &msg[1];
	}

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			bucketlen + strlen(msg) + 32);
	format_message(output_buffer, "%d %s %s\n", CODE_LOG, bucket, msg);
	log_write(output_buffer);

out_of_memory:
	return flush_output();
}


int send_missing(const char *cmd, const char *tag)
{
	return send_msg("ERROR", "ReqMissing", "ss", cmd, tag);
}


int send_msg(const char *bucket, const char *msgkey, const char *fmt, ...)
{
	va_list va;
	char *s;
	size_t len;

	va_start(va, fmt);
	s = buffer_format_v(&len, bucket, msgkey, fmt, va);

	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			len + 16);
	memcpy(output_buffer, s, len + 1);

out_of_memory:
	free(s);
	va_end(va);
	return flush_output();
}


/*
 *  function to output a result block 
 */
int send_result(int rc)
{
	BUFFER_ENSURE_CAPACITY(
			&output_buffer,
			&output_buffer_size,
			64);
	format_message(output_buffer, "%d RESULT %d\n%d EOR\n",
			CODE_RESULT,
			(rc >= 0) ? rc : -rc,
			CODE_EOR);

out_of_memory:
	/* dump the output cache if there's data in it */
	if (!flush_output()) { return 0; }
	return send_cache_flush();
}


int send_script(const char *dir, char *cmd, int verbose) {
	char *p = cmd;
	char *nl;

	if (verbose) {
		while (NULL != ( nl = strchr(p, '\n') )) {
			char c = *nl;
			*nl = '\0';
			if (!send_log("SCRIPT", p)) { return 0; }
			*nl = c;
			p = &nl[1];
		}
		if (!send_log("SCRIPT", p)) { return 0; }
	}
	return send_msg("EXEC", "ExecStart", "ss",
			dir, platform_hostname());
}


